/*-
 * Copyright (C)2018..2020 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2018..2020 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: engine.c,v 1.28 2020/05/16 01:01:00 babolo Exp $"

#define MIFE_COMPAT     5
#define BLIN_COMPAT     3
#define RECOBE_COMPAT   4
#define RECOBE_INTERNAL 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <mife.h>
#include "recobe.h"

static recobe_chunk *
recobe_getmife(recobe_chain *chain) {
    recobe_chunk    *chunk = NULL;
    off_t            toff = 0;
    ssize_t          insz;
    u_char          *inin;
    mife_descriptor *md;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    RECOBE_CHAIN(chain);
    if  (!chain->converdata) {
        ifBLIN_QX0("No converdata");
        errno = EFAULT;
        goto out;
    }
    md = chain->converdata;
    if  (chain->flags & RECOBE_USED) toff = md->ofpoint + chain->outblock;
    ifBLIN_QX3("  toff=%"BLIN_O"d", toff);
    if  (!(inin = mife_window(md, toff, (ssize_t)chain->inblock))) {
        ifBLIN_QW0("mife_window(,%"BLIN_D",%"BLIN_O"d)", toff, (ssize_t)chain->inblock);
        goto out;
    }
    if  (0 > (insz = mife_ctlsize(md))) {
        ifBLIN_QW0("mife_ctlsize");
        goto out;
    } else if (insz > (ssize_t)chain->inblock) {
        insz = (ssize_t)chain->inblock;
    }
    chain->outblock = insz;
    chain->flags |= RECOBE_USED;
    if  (!(chunk = malloc(insz + sizeof(recobe_chunk)))) {
        ifBLIN_QW0("No mem %d", insz + sizeof(recobe_chunk));
        goto out;
    }
    chunk->length = insz;
    chunk->position = !insz ? RECOBE_EOT : 0;
    bcopy(inin, chunk->chunk, chunk->length);
out:
    RECOBE_CHUNK(chunk);
    ifBLIN_QX3("- %"BLIN_X, chunk);
    return(chunk);
#   undef blin_internal_flags
}

recobe_chain *
recobe_openmd(recobe_conf *cf, mife_descriptor *md, u_int32_t chunksize) {
#   define blin_internal_flags (cf->flags)
    recobe_chain *chain = NULL;

    ifBLIN_QX3("+ [%d]", chunksize);
    if  (!(chain = malloc(sizeof(recobe_chain)))) {
        ifBLIN_QW0("malloc(%d)", sizeof(recobe_chain));
        goto out;
    }
    chain->converdata = md;
    chain->cf = cf;
    chain->flags = (cf->flags & ~RECOBE_USED) | RECOBE_CHAINER;
    chain->tail = NULL;
    chain->next = NULL;
    chain->outblock = 0;
    chain->inblock = chunksize;
    chain->puller = recobe_getmife;
    chain->cnv = (recobe_conv){0, 0, 1, 0, 1};
out:
    RECOBE_CHAIN(chain);
    ifBLIN_QX3("- %"BLIN_X, chain);
    return(chain);
#   undef blin_internal_flags
}

recobe_chain *
recobe_openfile(recobe_conf *cf, const char *filename, u_int32_t chunksize) {
#   define blin_internal_flags (cf->flags)
    recobe_chain    *chain = NULL;
    mife_descriptor *md = NULL;

    ifBLIN_QX3("+ %s[%d]", filename, chunksize);
    if  (!(md = mife_init((blin_internal_flags & BLIN_GEN1) | MIFE_SLID))) {
        ifBLIN_QW0("mife_init");
        goto out;
    }
    if  (!!(filename)) {
        if  (0 > mife_ctlfile(md, filename)) {
            ifBLIN_QW0("mife_ctlfile(,%s)", filename);
            goto out;
        }
    } else {
        if  (!!fflush(stdin)) ifBLIN_QW1("fflush");
        if  (0 > mife_ctlfdsc(md, fileno(stdin))) {
            ifBLIN_QW0("mife_ctlfdsc(,%d)", fileno(stdin));
            goto out;
    }   }
    chain = recobe_openmd(cf, md, chunksize);
out:
    ifBLIN_QX3("- %"BLIN_X, chain);
    return(chain);
#   undef blin_internal_flags
}

int
recobe_inword(recobe_chain *chain, recobe_chunk *chunk, u_int64_t *wchr) {
    u_int32_t     l = (chain->inwlen > 8) ? 8 : chain->inwlen;
    u_char       *c = &chunk->chunk[chunk->position];
    int           ex = 0;

#   define blin_internal_flags (chain->flags)
    ifBLIN_QX3("+ %08X %"BLIN_X, chain->flags, BLIN_I(chain));
    if  (!recobe_chunksize(chunk)) goto out;
    if  (chunk->position + l > chunk->length) {
        ex = -(int)recobe_chunksize(chunk);
        goto out;
    }
    if  (!(chain->flags & RECOBE_ORDER)) {
        *wchr = recobe_load(&c[ex], l);
        ex += l;
    }
    if  ((8 < chain->inwlen) && !!c[ex++]) {
        ex = -ex;
        goto out;
    }
    if  ((9 < chain->inwlen) && !!c[ex++]) {
        ex = -ex;
        goto out;
    }
    if  (!!(chain->flags & RECOBE_ORDER)) {
        *wchr = recobe_load(&c[ex], -l);
        ex += l;
    }
out:;
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

void
recobe_outword(recobe_chain *chain, recobe_chunk *chunk, u_int64_t wchr) {
    recobe_store( &chunk->chunk[chunk->length]
                , (chain->flags & RECOBE_ORDER) ? -(chain->outwlen) : chain->outwlen
                , wchr
                )
    ;
    chunk->length += chain->outwlen;
}

static void
toer(recobe_chain *chain, recobe_chunk *chunk) {
    for (int k = 0; !!chain->bsubst && !!chain->bsubst[k]; ++k) {
        recobe_outword(chain, chunk, chain->bsubst[k]);
    }
    for (int k = 0; !!chain->esubst && !!chain->esubst[k]; ++k) {
        recobe_outword(chain, chunk, chain->esubst[k]);
}   }

static recobe_chunk *
convey(recobe_chain *chain) {
    recobe_chunk *chunk = NULL;
    u_int32_t     tailen = 0;
    recobe_chunk *source;
    int           ex = EX_OK;

    ifBLIN_QX3("+ chain=%"BLIN_X, chain);
    if  (!chain) {
        ifBLIN_QW0("No chain");
        errno = EFAULT;
        goto out;
    }
#   define blin_internal_flags (chain->flags)
    RECOBE_CHAIN(chain);
    if  (chain->flags & RECOBE_CHAINER) {
        chunk = chain->puller(chain);
        goto out;
    }
    for (                          /*          */
        ; ((u_int32_t)chain->inwlen * chain->inwnum) > (tailen = recobe_chunksize(chain->tail))
        ;
        ) {
        RECOBE_CHAIN(chain);
        if  (!!chain->next) {                                              /*    */
            if  (!(source = convey(chain->next))) {
                ifBLIN_QW0("convey");
                goto out;
            }
            RECOBE_CHUNK(chunk);
            if  (0 > (ex = recobe_chunksum(chain, source))) {
                ifBLIN_QW0("recobe_chunksum");
                free(source);
                source = NULL;
                goto out;
            }
            free(source);
            source = NULL;
            if  (1 == ex) break;
            if  (2 == ex) {
                chunk = recobe_chunkeot(chain);
                goto out;
            }
        } else if (!tailen) {                                        /*      */
            chunk = recobe_chunkeot(chain);
            goto out;
    }   }
    RECOBE_CHAIN(chain);
    tailen = recobe_chunksize(chain->tail);           /*        */
    ifBLIN_QX4("tailen=%u", tailen);
    if  (!!chain->convertor) {                        /*       */
        u_int32_t     forer = 0;

        tailen = tailen * chain->outwlen * chain->outwnum / chain->inwlen + 1;
        if  (!!chain->bsubst) forer += strlen((const char*)chain->bsubst);
        if  (!!chain->esubst) forer += strlen((const char*)chain->esubst);
        forer *= chain->outwlen;
        if  (!!tailen && (tailen < forer)) tailen = forer;
        if  (!(chunk = malloc(tailen + sizeof(recobe_chunk)))) {
            ifBLIN_QW0("malloc(%d)", tailen + sizeof(recobe_chunk));
            goto out;
        }
        chunk->length = 0;
        chunk->position = 0;
        if  (!chain->tail) goto out;
        while(!!recobe_chunksize(chain->tail)) {
            int           ex;

            ex = chain->convertor(chain, chunk);
            if  (0 > ex) toer(chain, chunk);
            if  (  !ex
                || (tailen < chunk->length + forer)
                || (tailen < chunk->length + (chain->outwnum * chain->outwlen))
                || (chain->tail->length < chain->tail->position + (chain->inwnum * chain->inwlen))
                ) {
                break;
        }   }
    } else if (chain->tail) {
        ifBLIN_QX0("XXXX DEAD CODE 1 XXXX");
        if  (RECOBE_PREFLEN < tailen) {
            if  (!(chunk = malloc(sizeof(recobe_chunk) + RECOBE_PREFLEN))) {
                ifBLIN_QW0("malloc(%"BLIN_D")", sizeof(recobe_chunk) + RECOBE_PREFLEN);
                goto out;
            }
            memcpy(chunk->chunk, &chain->tail[chain->tail->position], RECOBE_PREFLEN);
            chain->tail->position += RECOBE_PREFLEN;
            chunk->length = RECOBE_PREFLEN;
            chunk->position = 0;
        } else {
            chunk = chain->tail;
            chain->tail = NULL;
        }
    } else {
        ifBLIN_QX0("XXXX DEAD CODE 2 XXXX");
        if  (!(chunk = malloc(sizeof(recobe_chunk)))) {
            ifBLIN_QW0("malloc(%"BLIN_D")", sizeof(recobe_chunk));
            goto out;
        }
        bzero(chunk, sizeof(recobe_chunk));
    }
out:
    if  ((chain->flags & RECOBE_TRACE) && !!chunk) {
        const recobe_c_tnm *encolist = recobe_collection.domains[chain->cnv.domain]->encolist;

        fprintf( stderr
               , "@ %s=%c%c(%d: %c%d => %c%d) %"BLIN_X" %u[%u]\n"
               , recobe_collection.domains[chain->cnv.domain]->domname
               , chain->cnv.nonesrc ? '.' : encolist[chain->cnv.encosrc].encoid[0]
               , chain->cnv.nonedst ? '.' : encolist[chain->cnv.encodst].encoid[0]
               , chain->cnv.domain
               , chain->cnv.nonesrc ? '-' : '+'
               , chain->cnv.encosrc
               , chain->cnv.nonedst ? '-' : '+'
               , chain->cnv.encodst
               , BLIN_I(chunk)
               , chunk->position
               , chunk->length
               )
        ;
        fprintf(stderr, "");
        for (u_int32_t i = chunk->position; i < chunk->length; ++i) {
            if  ((chunk->chunk[i] < ' ') || ((chunk->chunk[i] >= 0x7F) && (chunk->chunk[i] < 0xC0))) {
                fprintf(stderr, "\\%02X", chunk->chunk[i]);
            } else {
                fprintf(stderr, "%c", chunk->chunk[i]);
        }   }
        fprintf(stderr, "\n");
    }
    RECOBE_CHUNK(chunk);
    ifBLIN_QX3("- %"BLIN_X, chunk);
    return(chunk);
#   undef blin_internal_flags
}

int
recobe_chainput(recobe_chain *chain, int fd, ssize_t (*wrt)(int fd, const void *buf, size_t size)) {
    recobe_chunk *toout = NULL;
    int           ex = EX_OK;
    u_int32_t     length;
    ssize_t       wes;

    ifBLIN_QX3("+ chain=%"BLIN_X" to %d", chain, fd);
    if  (!chain) {
        ifBLIN_QW0("No chain");
        ex = EX_NOINPUT;
        errno = EFAULT;
        goto out;
    }
#   define blin_internal_flags (chain->flags)
    RECOBE_CHAIN(chain);
    for (int i = 0;; ++i) {
        if  (!(toout = convey(chain))) {
            recobe_cnverr(__FILE__, __func__, __LINE__, "convey", chain->cnv);
            ex = EX_SOFTWARE;
            goto out;
        }
        length = recobe_chunksize(toout);
        wes = (0 >= length) ? 0 : wrt(fd, &toout->chunk[toout->position], length);
        if  (RECOBE_EOT == toout->position) {
            free(toout);
            toout = NULL;
            break;
        }
        free(toout);
        toout = NULL;
        if  (0 > wes) {
            ifBLIN_QW0("mife_writ");
            ex = EX_IOERR;
            goto out;
        }
        if  ((ssize_t)length != wes) {
            ifBLIN_QW0("mife_writ(,,%d) <> %d", (int)length, (int)wes);
            ex = EX_IOERR;
            goto out;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
recobe_chainout(recobe_chain *chain, int fd) {
    return(recobe_chainput(chain, fd, mife_writ));
}
